PCI_CONFIG_ADDRESS equ 0xCF8
PCI_CONFIG_DATA    equ 0xCFC

XHCI_CLASS      equ 0x0C
XHCI_SUBCLASS   equ 0x03
XHCI_PROGIF     equ 0x30

BLACK         equ 0x0
BLUE          equ 0x1
GREEN         equ 0x2
CYAN          equ 0x3
RED           equ 0x4
MAGENTA       equ 0x5
BROWN         equ 0x6
LIGHT_GREY    equ 0x7
DARK_GREY     equ 0x8
LIGHT_BLUE    equ 0x9
LIGHT_GREEN   equ 0xA
LIGHT_CYAN    equ 0xB
LIGHT_RED     equ 0xC
LIGHT_MAGENTA equ 0xD
LIGHT_BROWN   equ 0xE
WHITE         equ 0xF

%macro print 1+
    [section .data]
%%string:
    db %1,0
    __SECT__
    mov si,%%string
    call print_str
%endmacro



org 0x7E00
bits 16

section .text

cli
lgdt [global_descriptor_table]
mov	eax, cr0
or eax, 1
mov	cr0, eax
jmp 8*1:protected_mode



bits 32

protected_mode:
mov edi, 0
mov ax, 8*2
mov ds, ax
mov ss, ax
mov es, ax
mov fs, ax
mov gs, ax
mov esp, 0x0000_0500

mov eax, XHCI_CLASS << 8*3 | XHCI_SUBCLASS << 8*2 | XHCI_PROGIF << 8*1 | 0x00 << 8*0
call find_pci_device

call get_sbrn_register

mov cl, 2
mov dh, BLUE << 4 | GREEN
print "SBRN: "
call print_hex_digits

cli
hlt
jmp $



section .data

global_descriptor_table:

; GDT descriptor
dw 40 - 1
dd $ + 4

; Null
dq 0

; Code
dw 0xFFFF
dw 0x0000
db 0x00
db 0b1____00____1____1____0____1____0
db 0b1____1____00 << 4 | 0xF
db 0x00

; Data
dw 0xFFFF
dw 0x0000
db 0x00
db 0b1____00____1____0____0____1____0
db 0b1____1____00 << 4 | 0xF
db 0x00

; Video
dw 0xFFFF
dw 0x8000
db 0x0B
db 0b1____00____1____0____0____1____0
db 0b1____1____00 << 4 | 0xF
db 0x00


section .functions


; mov esi, string           ; String to print
; mov edi, 0x04             ; Where to print in screen
; mov dh, BLUE << 4 | GREEN ; Colour

; call print_str

print_str:
    push dx
    push esi
    push es
    push ax
    ; DI is not pushed/poped so you can
    ; get the last printed location to
    ; keep printing more easily
    mov ax, 8*3
    mov es, ax
.print:
    ; dl contains each character
    mov dl, [esi]
    cmp dl, 0
    jz .end
    inc esi
    mov [es:edi*2], dx
    inc edi
    jmp .print
.end:
    pop ax
    pop es
    pop esi
    pop dx
    ret


; mov dl, 0x0C              ; Hexadecimal number to print
; mov edi, 0x04             ; Where to print in screen
; mov dh, BLUE << 4 | GREEN ; Colour

; call print_hex

print_hex:
    push dx
    push es
    push ax
    ; DI is not pushed/poped so you can
    ; get the last printed location to
    ; keep printing more easily
    mov ax, 8*3
    mov es, ax
    and dl, 0x0F
    ; Check if we must print a number [0-9] or a letter [A-F]
    cmp dl, 0x0A
    jge .is_letter
    add dl, '0'
    jmp .end_adding
.is_letter:
    add dl, 'A'-0x0A
.end_adding:
    mov [es:edi*2], dx
    inc edi
    pop ax
    pop es
    pop dx
    ret


; mov eax, 0x12345678       ; Hexadecimal number to print
; mov cl, 0x2               ; Number of heaxedecimal digits to print
; mov edi, 0x04             ; Where to print in screen
; mov dh, BLUE << 4 | GREEN ; Colour

; call print_hex_digits

print_hex_digits:
    push eax
    push ebx
    push cx
    push dx
    ; DI is not pushed/poped so you can
    ; get the last printed location to
    ; keep printing more easily
    shl cl, 2
    jz .end
    mov ebx, eax
    mov ax, 8*3
    mov es, ax
.loop:
    mov eax, ebx
    sub cl, 4
    shr eax, cl
    mov dl, al
    call print_hex
    cmp cl, 0
    je .end
    push cx
    and cl, 0b0000_0111
    cmp cl, 0b0000_0000
    jne .next
    mov dl, '_'
    mov [es:edi*2], dx
    inc edi
.next:
    pop cx
    jmp .loop
.end:
    pop dx
    pop cx
    pop ebx
    pop eax
    ret


; Enable Bit    Reserved    Bus Number    Device Number    Function Number    Register Offset
; 31            30 - 24     23 - 16       15 - 11          10 - 8             7 - 0
; mov eax, 0b1_0000000_00000000_00000_000_00000000 ; Where to read the data from
; call read_32_bits_pci_register

read_32_bits_pci_register:
    push dx
    mov dx, PCI_CONFIG_ADDRESS
    out dx, eax
    mov dx, PCI_CONFIG_DATA
    in eax, dx
    pop dx
    ret


; mov eax, XHCI_CLASS << 24 | XHCI_SUBCLASS << 16 | XHCI_PROGIF << 8 | 0x00
; call find_pci_device

find_pci_device:
    push ebx
    push ecx
    push dx
    mov ecx, eax
    ; Enable Bit    Reserved    Bus Number    Device Number    Function Number    Register Offset
    ; 31            30 - 24     23 - 16       15 - 11          10 - 8             7 - 0
    mov eax, 0b1_0000000_00000000_00000_000_00000000
.next:
    mov ebx, eax
    ; Check if there is a device
    call read_32_bits_pci_register
    cmp eax, 0xFFFF_FFFF
    je .end_check
    mov eax, ebx
    mov al, 8
    call read_32_bits_pci_register
    and eax, 0xFF_FF_FF_00
    cmp eax, ecx
    je .end
.end_check:
    mov eax, ebx
    add eax, 0b0_0000000_00000000_00000_001_00000000
    cmp eax, 0b1_0000001_00000000_00000_000_00000000
    jne .next
    ; Device not found
    jmp $
.end:
    mov eax, ebx
    pop dx
    pop ecx
    pop ebx
    ret


get_sbrn_register:
    mov al, 0x60
    call read_32_bits_pci_register
    ret
